(**************************************************************************)
(*                                                                        *)
(*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       *)
(*  Kim Nguyen, Laurent Sartre                                            *)
(*                                                                        *)
(*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      *)
(*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 *)
(*                                                                        *)
(*  https://www.informatique-mpi.fr/                                      *)
(*                                                                        *)
(**************************************************************************)

(*
  Lempel--Ziv--Welch

  https://en.wikipedia.org/wiki/Lempel%E2%80%93Ziv%E2%80%93Welch
*)

let debug = false

type compressed = string

(* - l'alphabet est 0,1 uniquement ; ce serait les bits du texte
     à compresser
   - les codes émis ont une longueur variable, qui commence à 1 et grandit
     au fur et à mesure de la compression
*)

(* écrit `c` sur `n` chiffres en binaire, dans le buffer `b` *)
let rec output (b: Buffer.t) (n: int) (c: int) : unit =
  if n > 0 then (
    output b (n - 1) (c / 2);
    Buffer.add_char b (if c mod 2 = 1 then '1' else '0')
  )

let encode (s: string) : string =
  let n = String.length s in
  let codes = Trie.create () in
  Trie.put codes "0" 0;
  Trie.put codes "1" 1;
  let size = ref 1 in
  let next = ref 2 in
  let b = Buffer.create 1024 in
  let i = ref 0 in
  while !i < n do
    let p, c = Trie.prefix_sub codes s !i in
    output b !size c;
    if debug then Format.eprintf "  out %d (%d bits)@." c !size;
    if !i + p < n then (
      let w = String.sub s !i (p + 1) in
      Trie.put codes w !next;
      if debug then Format.eprintf "  %d -> %S@." !next w;
      if !next = 1 lsl !size then incr size;
      incr next
    );
    i := !i + p
  done;
  (* if debug then Format.eprintf "  next = %d / size = %d@." !next !size; *)
  Buffer.contents b

(* lit `size` bits au début de `s[i..[` *)
let input (size: int) (s: string) (i: int) : int =
  assert (i + size <= String.length s);
  int_of_string ("0b" ^ String.sub s i size)

let decode (s: string) : string =
  let words = Vector.create 2 "" in
  Vector.push words "0";
  Vector.push words "1";
  let size = ref 1 in
  let b = Buffer.create 1024 in
  let i = ref 0 in
  let last = ref "" in
  while !i < String.length s do
    let c = input !size s !i in
    i := !i + !size;
    if debug then Format.eprintf "  input %d (%d bits)@." c !size;
    if !last <> "" then (
      let next = Vector.size words in
      let w = if c = next then !last else Vector.get words c in
      Vector.push words (!last ^ String.sub w 0 1);
      if debug then
        Format.eprintf "  %d -> %S@." next (Vector.get words next);
    );
    if Vector.size words = 1 lsl !size then incr size;
    last := Vector.get words c;
    Buffer.add_string b !last
  done;
  Buffer.contents b


(* Dans variante, la taille des codes est fixe et passée
   en argument (`size`).

   Le nombre de codes est donc limité à 2^size (`maxc` ci-dessous).
*)

module FixedSize = struct

  let one c = String.make 1 (Char.chr c)

  let encode size s =
    assert (size >= 8);
    let n = String.length s in
    let maxc = 1 lsl size in
    let codes = Trie.create () in
    for c = 0 to 255 do Trie.put codes (one c) c done;
    let next = ref 256 in
    let b = Buffer.create 1024 in
    let rec encode i =
      if i < n then (
        let p, c = Trie.prefix_sub codes s i in
        output b size c;
        if i + p < n && !next < maxc then (
          let w = String.sub s i (p + 1) in
          Trie.put codes w !next;
          if debug then Format.eprintf "  %d -> %S@." !next w;
          incr next
        );
        encode (i + p)
      ) in
    encode 0;
    Buffer.contents b

  let decode size s =
    let maxc = 1 lsl size in
    let words = Array.init maxc (fun i -> if i < 256 then one i else "") in
    let b = Buffer.create 1024 in
    let next = ref 256 in
    let rec decode last i =
      if i <> String.length s then (
        let c = input size s i in
        assert (c <= !next);
        if last <> "" && !next < maxc then (
          let w = if c = !next then last else words.(c) in
          words.(!next) <- last ^ String.sub w 0 1;
          if debug then Format.eprintf "  %d -> %S@." !next words.(!next);
          incr next
        );
        Buffer.add_string b words.(c);
        decode words.(c) (i + size)
      ) in
    decode "" 0;
    Buffer.contents b

  (* signale les occurrences de `m` pendant la décompression
     (sans stocker l'intégralité du texte décompressé) *)
  let decode_and_search size s m =
    let maxc = 1 lsl size in
    let words = Array.init maxc (fun i -> if i < 256 then one i else "") in
    let b = Rabin_karp.Buffer.create m in
    let next = ref 256 in
    let rec decode last i =
      if i <> String.length s then (
        let c = input size s i in
        assert (c <= !next);
        if last <> "" && !next < maxc then (
          let w = if c = !next then last else words.(c) in
          words.(!next) <- last ^ String.sub w 0 1;
          if debug then Format.eprintf "  %d -> %S@." !next words.(!next);
          incr next
        );
        Rabin_karp.Buffer.add_string b words.(c);
        decode words.(c) (i + size)
      ) in
    decode "" 0;
    Rabin_karp.Buffer.count b

end
